--***************************************************************************
-- Copyright (c) 1994 - Ohio Board of Regents and
--                    - University of Cincinnati, Cincinnati, OH.
--                    - Sidhartha Mohanty,
--                    - Department of Electrical & Computer Engineering,
--                    - Jan, 1994.
--***************************************************************************
--
-- Title:               Design library for Queueing Model. Version 1.0.
--
-- File:                [SUITE: vhdl/project/queue/queue4] server.v
-- Created:             Fri Sept. 16 19:19:26 1994
-- Author:              Sidhartha Mohanty <smohanty@thor.ece.uc.edu>
--
-- Keywords:
-- LRM_version:         IEEE Std. P1076-1992/C
--
-- Objective:
--
-- Usage: VHDL Design Entity for Server module.  Analyze after rand.vhdl.
-- 
-- Algorithm :
-- (None)
--
-- References:
--
-- Result:
--
-- History: 
--
-- Comments: 
--
--***************************************************************************

library queue;
use queue.queue_pkg.all;
use queue.rng.all;
use std.textio.all;

entity service is
 
 generic ( service_time     : time         := 1 fs;
	   name 	    : string       := "S2";
	   service_dist     : distrib      := unif;
           server_type      : service_type := single_token;
	   service_capacity : integer      := 1;  
	   seed		    : integer      := 1;
	   generation       : frequency    := (1.0, 1.0);
	   debug_control    : debug        := none;
	   file_name 	    : string	   := "";
	   report_freq      : time         := 10 fs );

 port (in_arc : inout arc bus; out_arc : inout arc bus);

end;

architecture behavior of service is

  type squeue_type is
   record
     tok    :  token;
     tstamp : time;
   end record;

  type server_buff is array (natural range <>) of squeue_type;
  signal service_queue   : server_buff(1 to service_capacity);
  signal tail 	         : integer  := 1;
  signal head            : integer  := 1;
  signal change          : tail_ptr := 0;
  signal number_serviced : integer := 0;
  signal number_changed  : tokens_change := 0;
  signal in_service      : integer := 0;
  signal serv_time	 : time := 0 fs;
  signal id 		 : integer := 1;
  file outfile : text is out file_name;

begin

initialize : process		-- Initialization Process.  Attach unique 
				-- identifications to each server module

   variable integer_line : line;
   variable integer_id   : integer := 1;

  begin

    in_arc <= null;
    out_arc <= null;

    assert not std.textio.endfile(std.textio.input)
	report "(server.vhdl) Number of entities in Queueing Model exceed number of integer ids. Increase the number."
	severity failure;

    readline(std.textio.input, integer_line);

    assert integer_line'length >= 0
	report "(server.vhdl) Missing integer in identifiers file."
	severity failure;

    read(integer_line, integer_id);
    id <= integer_id;

    wait for 0 fs;

    wait;

end process;

server : process			-- Main process for Server module.
					-- Initialization of Random variables.
   variable unf   : Uniform := InitUniform(seed, generation.mean, generation.var);
   variable norml : Normal  := InitNormal(seed,generation.mean,generation.var);
   variable pois  : Poisson := InitPoisson(seed, generation.mean);
   variable erlan : Erlang  := InitErlang(seed,generation.mean,generation.var);
   variable rnd   : Real    := 1.0;
   variable servc_time    : time := 1 fs;
   variable time_stamp      : time := 0 fs;
   variable temp, temp0     : squeue_type := ((0, ready, FALSE, 0 fs), 0 fs);
   variable index 	    : integer := 0;
   variable mesg 	    : line;

 begin

   in_arc <= null;
   out_arc <= null;
   
   loop
    wait  until (in_arc.handshake = sending and  in_service < service_capacity); 
    if (server_type = single_token) then
       write(mesg, string'("Server: "));
       write(mesg, name);
       write(mesg, string'(" is of type: "));
       write(mesg, string'("single_token "));
       write(mesg, string'(" service_capacity = "));
       write(mesg, service_capacity);
       writeline(outfile, mesg);

	assert (service_capacity = 1)
	  report "Error in service_capacity specifications"
	  severity error;
    elsif (server_type = multi_token) then
       write(mesg, string'("Server: "));
       write(mesg, name);
       write(mesg, string'(" is of type: "));
       write(mesg, string'("multi_token "));
       write(mesg, string'(" service_capacity = "));
       write(mesg, service_capacity);
       writeline(outfile, mesg);

	  assert (service_capacity > 1)
	    report "Error in service_capacity specifications"
	    severity warning;
    end if;

    if (debug_control = handshaking or 
	debug_control = summary) then
    print_handshake_message ("Server: ", name, " received handshake signal: ", 
			     in_arc.handshake,
                             " from component: ", in_arc.id,
			     " at Time: ", now);
    end if;

   if (in_service < service_capacity) then

      if (debug_control = trace_tokens or 
	  debug_control = summary) then
      print_server_message ("Server ", name, " accepting tokens from queue: ",
                             in_arc.id, " for Service at Time: ", now);
      end if;

      if (service_dist = unif) then
            assert false
              report "Server with service rates following uniform distribution"
              severity note;
	    if (debug_control = trace_tokens or 
	        debug_control = summary) then

            write(mesg, string'("Server: "));
	    write(mesg, name);
	    write(mesg, string'(" with service rates following UNIFORM distribution; Lower & Upper Bounds: "));
	    write(mesg, generation.mean);
	    write(mesg, string'(" "));
	    write(mesg, generation.var);
	    writeline(outfile, mesg);
            print_source_message("Server: ", name, " with service rates following ",
                                  service_dist, "orm distribution; Lower & Upper bounds: ",
                                  generation.mean, generation.var); 
	    end if;
            GenRnd(unf);
            rnd := unf.rnd;
            servc_time := rnd * 1.0 fs;
      elsif (service_dist = norm) then
            assert false
              report "Server with service rates following normal distribution"
              severity note;
            if (debug_control = trace_tokens or 
	        debug_control = summary) then

	    write(mesg, string'("Server: "));
	    write(mesg, name);
	    write(mesg, string'(" with service rates following NORMAL distribution; Mean & Var: "));
	    write(mesg, generation.mean);
	    write(mesg, string'(" "));
	    write(mesg, generation.var);
	    writeline(outfile, mesg);

            print_source_message("Server: ", name, " with service rates following ",
                                  service_dist, "al distribution; Mean & Var: ",
                                  generation.mean, generation.var); 
	    end if;
            GenRnd(norml);
            rnd := norml.rnd;
            servc_time := rnd * 1.0 fs;
      elsif (service_dist = poiss) then
            assert false
              report "Server with service rates following poisson distribution"
              severity note;
	    if (debug_control = trace_tokens or 
	        debug_control = summary) then

	    write(mesg, string'("Server: "));
	    write(mesg, name);
	    write(mesg, string'(" with service rates following POISSON distribution; Mean: "));
	    write(mesg, generation.mean);
	    writeline(outfile, mesg);

            print_source_message("Server: ", name, " with service rates following ",
                                  service_dist, "an distribution; Mean & Var: ",
                                  generation.mean, generation.var); 
	    end if;
            GenRnd(pois);
            rnd := pois.rnd;
            servc_time := rnd * 1.0 fs;
      elsif (service_dist = erl) then
            assert false
              report "Server with service rates following binomial distribution"
              severity note;
	    if (debug_control = trace_tokens or 
	        debug_control = summary) then

	    write(mesg, string'("Server: "));
	    write(mesg, name);
	    write(mesg, string'(" with service rates following ERLANG distribution; Mean & Var: "));
	    write(mesg, generation.mean);
	    write(mesg, string'(" "));
	    write(mesg, generation.var);
	    writeline(outfile, mesg);

            print_source_message("Server: ", name, " with service rates following ",
                                  service_dist, "ang distribution; Mean & Var: ",
                                  generation.mean, generation.var); 
	    end if;
            GenRnd(erlan);
            rnd := erlan.rnd;
            servc_time := rnd * 1.0 fs;
      elsif (service_dist = fixed) then
            assert false
              report "Server with service rates following fixed distribution"
              severity note;
	    if (debug_control = trace_tokens or 
	        debug_control = summary) then

	    write(mesg, string'("Server: "));
	    write(mesg, name);
	    write(mesg, string'(" with fixed service intervals: "));
	    write(mesg, generation.mean);
	    writeline(outfile, mesg);
            print_source_message("Server: ", name, " with service rates following ",
                                  service_dist, " distribution; Mean & Var: ",
                                  generation.mean, generation.var); 
	    end if;
            servc_time := service_time; -- For General Dist., we use the default
		    			  -- generic defined service time.
      end if;

      serv_time <= service_time;
      wait for 0 fs;
      time_stamp := serv_time;

      temp0 := (in_arc, time_stamp);

      if (tail = service_capacity + 1) then 
	  change <= - service_capacity;
	  wait for 0 fs;
	  change <= 0;
	  wait for 1 fs;
      end if;

     if in_service = 0 then
         service_queue(tail) <= temp0;
         wait for 1 fs;
     else
	if (tail < head) then
	  for I in head to service_capacity loop
	    if temp0.tstamp < service_queue(I).tstamp then
	      temp := service_queue(i);
	      service_queue(i) <= temp0;
	      wait for 1 fs;
            else 
	      service_queue(tail) <= temp0;
	      wait for 1 fs;
            end if;
          end loop;
	  index := tail - 1;
	  for I in 1 to index loop
	    if temp0.tstamp < service_queue(I).tstamp then
	      temp := service_queue(i);
	      service_queue(i) <= temp0;
	      wait for 1 fs;
            else 
	      service_queue(tail) <= temp0;
	      wait for 1 fs;
            end if;
          end loop;
	else
         index := tail - 1;  
         for I in head to index loop
	   if temp0.tstamp < service_queue(I).tstamp then
	      temp := service_queue(i);
	      service_queue(i) <= temp0;
	      wait for 1 fs;
           else 
	      service_queue(tail) <= temp0;
	      wait for 1 fs;
           end if;
         end loop;
	end if;
     end if;

     change <= + 1;
     wait for 1 fs;
     change <= 0;
     wait for 0 fs;
     number_changed <= + 1;
     wait for 1 fs;
     number_changed <= 0;
     wait for 3 fs;
     if in_service < service_capacity then 
        in_arc <= (id, accepted, FALSE, now);
        wait for 0 fs;
     else wait until in_service < service_capacity;
	in_arc <= (id, accepted, FALSE, now);
        wait for 0 fs;
     end if;
     if (debug_control = handshaking or 
	 debug_control = summary) then
     print_handshake_message ("Server: ", name, " sent handshake signal: ", 
	 	               in_arc.handshake," to component: ", 
			       in_arc.id," at Time: ", now);
    	end if;
     in_arc <= null;
     wait for 0 fs;
       
   else assert false
	  report "Server full, wait till free"
	  severity warning;
   end if;

 end loop;

end process;

server_dequeue : process		-- Dequeue Process for Multi-token 
					-- Server.
 
  variable start_time, start_time2   : time := 0 fs;
  variable idle_time, idle_time1     : time := 0 fs;
  variable total_time, busy_time     : time := 0 fs;
  variable utilization 		     : real := 0.0;
  variable mesg : line;

 begin
      
    out_arc <= null;
    loop
    start_time := now;  
    if (in_service > 0) then
      if (head = service_capacity + 1) then head <= 1; wait for 0 fs;
                                                       end if;
        if (debug_control = trace_tokens or
	    debug_control = summary) then
	 write(mesg, string'("Service Time = "));
	 write(mesg, service_queue(head).tstamp);
	 writeline(outfile, mesg);
	end if;

      wait for service_queue(head).tstamp;

        out_arc <= service_queue(head).tok;
        
	wait for 0 fs;
	head <= head + 1;
	wait for 0 fs;
	number_changed <= - 1;
        wait for 0 fs;
	number_changed <= 0;
	wait for 0 fs;
      
	if (debug_control = trace_tokens or 
	   debug_control = summary) then
	print_server_message ("Server ", name, " completed service for token from queue: ",
        	               in_arc.id, " at Time: ", now);
    	end if;
    	number_serviced <= number_serviced + 1;
	wait for 0 fs;
	write(mesg, string'("Total tokens serviced by Server "));
	write(mesg, name);
	write(mesg, string'(" at Time: "));
	write(mesg, now);
	write(mesg, string'(" = "));
	write(mesg, number_serviced);
	writeline(outfile, mesg);
    	print_token_message("Total tokens serviced by Server ", 
			    " at Time: ", now, " = ", number_serviced);
    	if (debug_control = handshaking or 
	    debug_control = summary) then
    	print_handshake_message ("Server: ", name, " sending handshake signal: ", 
	  		          out_arc.handshake," to component: ", 
			          out_arc.id," at Time: ", now);
        end if;
        start_time2 := now;
        wait on out_arc until (out_arc.handshake = accepted);
        idle_time := (now - start_time2);
    	if (debug_control = handshaking or 
	    debug_control = summary) then
      	print_handshake_message ("Server: ", name, " received handshake signal: ", 
	  		          out_arc.handshake," from component: ", 
			          out_arc.id," at Time: ", now);
        end if;
     	out_arc <= null;
    	wait for 0 fs;

	total_time := ((now - start_time) + idle_time1);
		-- Idle_time1 has been added because this is the time
		-- the server has to wait till it has a token to serve,
		-- i.e., till in_service > 0;

        idle_time := idle_time + idle_time1;
	busy_time  := (total_time - idle_time);
	utilization := (real(busy_time/1 fs) / real(total_time/1 fs));
	if (debug_control = summary) then
	    write(mesg, string'("Utilization for Server: "));
	    write(mesg, name);
	    write(mesg, string'("; U = "));
	    write(mesg, utilization);
	    writeline(outfile, mesg);
            if (utilization = 0.0) then 
	        write(mesg, string'("Server Idle throughout time period waiting for token"));
		writeline(outfile, mesg);
	    end if;
	end if;
        total_time := 0 fs; idle_time := 0 fs; idle_time1 := 0 fs;
        busy_time := 0 fs; utilization := 0.0;

    else
        wait until (in_service > 0);
        idle_time1 := (now - start_time);
        end if;
	
 end loop;
end process; 

count : process			-- Process for calculating indices of server
				-- queue.
 begin
   loop
     wait on number_changed;
     in_service  <= in_service + number_changed;
   end loop;
end process;

count_tail : process		-- Process for calculating indices of server
				-- queue.
 begin
   loop
     wait on change;
     tail <= tail + change;  
   end loop;
end process;

stats : process			-- Process for Calculating Performance Stats.
  
  variable total_time, avg_serv_time : time := 0 fs;
  variable k : integer := 0;
  variable L : line;

 begin
  wait on serv_time;
  total_time := total_time + serv_time;
  k := k + 1;
  if k = 10 then
    avg_serv_time := total_time / 10;
    if (debug_control = summary) then
      write(L, string'("Average Service Time for 10 tokens, Si = "));
      write(L, now);
      writeline(outfile, L);
      writeline(std.textio.output, L);
    end if;
    k := 0;
    total_time := 0 fs; 
  end if;
 end process;

stats1: process			-- Process for Calculating Performance Stats.
  variable L : line;
  variable throughput, temp : integer := 0;
 begin
   temp := number_serviced;   
   wait for 10 fs;
   throughput := (number_serviced - temp);
   if (debug_control = summary) then
      write(L, string'("Throughput of Server: "));
      write(L, name);
      write(L, string'(" over a period of 10 fs: "));
      write(L, now);
      write(L, string'("--"));
      write(L, (now - 10 fs));
      write(L, string'(" N = "));
      write(L, throughput);
      write(L, string'(" tokens/10 fs."));
      writeline(outfile, L);
      writeline(std.textio.output, L);
    end if;
 end process;

end behavior;
