--***************************************************************************
-- Copyright (c) 1994 - Ohio Board of Regents and
--                    - University of Cincinnati, Cincinnati, OH.
--                    - Sidhartha Mohanty,
--                    - Department of Electrical & Computer Engineering,
--                    - Sept, 1994.
--***************************************************************************
-- 
-- Title:               Design library for Queueing Model. Version 0.0.
-- 
-- File:                [SUITE: vhdl/project/queue] source.vhdl
-- Created:             Fri Sept 16, 10:14:45 1994
-- Author:              Sidhartha Mohanty <smohanty@thor.ece.uc.edu>
-- 
-- Keywords:
-- LRM_version:         IEEE Std. P1076-1992/C
-- 
-- Objective:
--
-- Usage:
-- Analyze after queue packages and random package.
--
-- Algorithm:               
-- (None)
-- 
-- References:
-- 
-- Result:
-- 
-- History:
-- 
-- Comments:
-- 
--***************************************************************************

library queue;
use queue.queue_pkg.all;
use queue.rng.all;
use std.textio.all;

entity source is

  generic (out_control   : distrib      := unif;
	   seed		 : integer 	:= 1;
	   generation    : frequency    := (50.0, 100.0);
	   name          : string	:= "S1";
	   debug_control : debug        := none;
	   file_name 	 : string	:= ""; 
	   report_freq   : time 	:= 10 fs);

  port (out_arc : inout arc bus);
    
end source;

architecture behavior of source is

  signal id : integer := 1;

begin

initialize : process

   variable integer_line : line;
   variable integer_id   : integer := 1;

  begin

    out_arc <= null;

    assert not std.textio.endfile(std.textio.input)
	report "(source.vhdl) Number of entities in Queueing Model exceed number of integer ids. Increase the number."
	severity failure;

    readline(std.textio.input, integer_line);

    assert integer_line'length >= 0
	report "(source.vhdl) Missing integer in identifiers file."
	severity failure;

    read(integer_line, integer_id);
    id <= integer_id;

    wait for 0 fs;

    wait;

end process;

source : process

   variable unf   : Uniform := InitUniform(seed, generation.mean, generation.var);
   variable norml : Normal  := InitNormal(seed, generation.mean, generation.var);
   variable pois  : Poisson := InitPoisson(seed, generation.mean);
   variable erlan : Erlang  := InitErlang(seed, generation.mean, generation.var);
   variable nex   : NegExp  := InitNegExp(seed, generation.mean);
   variable expon : Expn    := InitExpn(seed, generation.mean);
   variable rnd   : Real    := 1.0;
   variable trand : time    := 0.0 fs; 
   variable num_tok_gen     : integer := 0;
   variable mesg	    : line;
   variable start_time, idle_time : time := 0 fs;
   variable total_time, busy_time : time := 0 fs;
   variable utilization     : real := 0.0;
   file outfile : text is out file_name;

begin
    out_arc <= null;
    wait for 0 fs;
    loop
     num_tok_gen := num_tok_gen + 1;
    	assert out_arc.handshake /= not_ready
    	    report "Source overflow, output arc unable to accept new event."
    	    severity failure;
        if (out_control = unif) then
          assert false
            report "Generating tokens with uniform distribution"
            severity note;
          start_time := now;  
  	  if (debug_control = trace_tokens or 
  	      debug_control = summary) then 
	    write(mesg, string'("Source: "));
	    write(mesg, name);
	    write(mesg, string'(" generating tokens with UNIFORM distribution; LowerBound & UpperBound: "));
	    write(mesg, generation.mean);
	    write(mesg, string'(" "));
	    write(mesg, generation.var);
	    writeline(outfile, mesg);

-------------------------------------------------------------------------------
--          print_source_message("Source: ", name, " generating tokens with ",
--                 out_control, "orm distribution; LowerBound & UpperBound: ",
--                 generation.mean, generation.var); 
-------------------------------------------------------------------------------

          end if;
	  out_arc <= (id, sending, TRUE, now);

  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
          write(mesg, string'("Token generated at time: "));
	  write(mesg, now);
	  writeline(outfile, mesg);
	  write(mesg, string'("Total tokens generated by Source: "));
	  write(mesg, name);
	  write(mesg, string'(" at Time: "));
	  write(mesg, now);
	  write(mesg, string'(" = "));
	  write(mesg, num_tok_gen);
	  writeline(outfile, mesg);

  	  print_formatted_message("Token generated at time: ", now);
          print_token_message("Total tokens generated by Source ", 
  			      " at Time: ", now, " = ", num_tok_gen);
  	  print_handshake_message("Source: ", name, " sent handshake signal ",
               			   out_arc.handshake, " to Queue: ",
                                   out_arc.id, " at Time: ", now);
  	  end if;
          wait on out_arc until (out_arc.handshake = accepted);
          idle_time := (now - start_time);
  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
  	  print_handshake_message("Source: ", name, " received handshake signal ", 
                                   out_arc.handshake,
                                  " from component: ", out_arc.id,
                                  " at Time: ", now);
  	  end if;
          out_arc <= null;
          wait for 2 fs;
          idle_time := (idle_time + 2 fs);
          GenRnd(unf);
          rnd := unf.rnd;
          trand := rnd * 1.0 fs;
          wait for trand;
          idle_time := (idle_time + trand);
          total_time := (now - start_time);
          busy_time := (total_time - idle_time);
	  utilization := (real(busy_time/1 fs) / real(total_time/1 fs));
          if (debug_control = summary) then
	    write(mesg, string'("Utilization for Source: "));
	    write(mesg, name);
	    write(mesg, string'("; U = "));
	    write(mesg, utilization);
	    writeline(outfile, mesg);
	  end if;
	    
        elsif (out_control = norm) then
          assert false
            report "Generating tokens with normal distribution"
            severity note;
	  start_time := now;
  	  if (debug_control = trace_tokens or 
  	      debug_control = summary) then 
	    write(mesg, string'("Source: "));
	    write(mesg, name);
	    write(mesg, string'(" generating tokens with NORMAL distribution; Mean & Var: "));
	    write(mesg, generation.mean);
	    write(mesg, string'(" "));
	    write(mesg, generation.var);
	    writeline(outfile, mesg);

-------------------------------------------------------------------------------
--  	  print_source_message("Source: ", name, " generating tokens with ",
--                                out_control, "al distribution; Mean & Var: ",
--                                generation.mean, generation.var);
-------------------------------------------------------------------------------

  	  end if;
          out_arc <= (id, sending, TRUE, now);
  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
          write(mesg, string'("Token generated at time: "));
	  write(mesg, now);
	  writeline(outfile, mesg);
	  write(mesg, string'("Total tokens generated by Source: "));
	  write(mesg, name);
	  write(mesg, string'(" at Time: "));
	  write(mesg, now);
	  write(mesg, string'(" = "));
	  write(mesg, num_tok_gen);
	  writeline(outfile, mesg);

  	  print_formatted_message("Token generated at time: ", now);
  	  print_token_message("Total tokens generated by Source ", 
  			      " at Time: ", now, " = ", num_tok_gen);
  	  print_handshake_message("Source: ", name, " sent handshake signal ",
                 		   out_arc.handshake, " to Queue: ",
                                   out_arc.id, " at Time: ", now);
  	  end if;
          wait on out_arc until (out_arc.handshake = accepted);
	  idle_time := (now - start_time);
  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
  	  print_handshake_message("Source: ", name, " received handshake signal", 
                                   out_arc.handshake, " from component: ",
  			           out_arc.id," at Time: ", now);
  	  end if;
          out_arc <= null;
          wait for 2 fs;
	  idle_time := (idle_time + 2 fs);
          GenRnd(norml);
          rnd := norml.rnd;
          trand := rnd * 1.0 fs;
          wait for trand;
          total_time := (now - start_time);
          busy_time := (total_time - idle_time);
	  utilization := (real(busy_time/1 fs) / real(total_time/1 fs));
          if (debug_control = summary) then
	    write(mesg, string'("Utilization for Source: "));
	    write(mesg, name);
	    write(mesg, string'("; U = "));
	    write(mesg, utilization);
	    writeline(outfile, mesg);
	  end if;
	    
	elsif (out_control = poiss) then
	  assert false
            report "Generating tokens with poisson distribution"
            severity note;
	  start_time := now;
  	  if (debug_control = trace_tokens or 
  	      debug_control = summary) then 
	    write(mesg, string'("Source: "));
	    write(mesg, name);
	    write(mesg, string'(" generating tokens with POISSON distribution; Mean: "));
	    write(mesg, generation.mean);
	    writeline(outfile, mesg);

-------------------------------------------------------------------------------
--  	  print_source_message("Source: ", name, " generating tokens with ",
--                                out_control, "on distribution; Mean: ",
--                                generation.mean, generation.var);
-------------------------------------------------------------------------------

  	  end if;
          out_arc <= (id, sending, TRUE, now);

  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
          write(mesg, string'("Token generated at time: "));
	  write(mesg, now);
	  writeline(outfile, mesg);
	  write(mesg, string'("Total tokens generated by Source: "));
	  write(mesg, name);
	  write(mesg, string'(" at Time: "));
	  write(mesg, now);
	  write(mesg, string'(" = "));
	  write(mesg, num_tok_gen);
	  writeline(outfile, mesg);

  	  print_formatted_message("Token generated at time: ", now);
  	  print_token_message("Total tokens generated by Source ", 
  			      " at Time: ", now, " = ", num_tok_gen);
          print_handshake_message("Source: ", name, " sent handshake signal ",
          	     		   out_arc.handshake, " to Queue: ",
                                   out_arc.id, " at Time: ", now);
  	  end if;
	  wait on out_arc until (out_arc.handshake = accepted);
	  idle_time := (now - start_time);
  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
  	  print_handshake_message("Source: ", name,  " received handshake signal", 
  				   out_arc.handshake,
                                   " from component: ", out_arc.id,
                                   " at Time: ", now);
  	  end if;
          out_arc <= null;
          wait for 2 fs;
	  idle_time := (idle_time + 2 fs);
          GenRnd(pois);
          rnd := pois.rnd;
          trand := rnd * 1.0 fs;
          wait for trand;
          total_time := (now - start_time);
          busy_time := (total_time - idle_time);
	  utilization := (real(busy_time/1 fs) / real(total_time/1 fs));
          if (debug_control = summary) then
	    write(mesg, string'("Utilization for Source: "));
	    write(mesg, name);
	    write(mesg, string'("; U = "));
	    write(mesg, utilization);
	    writeline(outfile, mesg);
	  end if;
	    
   	elsif (out_control = erl) then
	  assert false
            report "Generating tokens with Erlang distribution"
            severity note;
	  start_time := now;
  	  if (debug_control = trace_tokens or 
  	      debug_control = summary) then 
	    write(mesg, string'("Source: "));
	    write(mesg, name);
	    write(mesg, string'(" generating tokens with ERLANG distribution; Mean & Var: "));
	    write(mesg, generation.mean);
	    write(mesg, string'(" "));
	    write(mesg, generation.var);
	    writeline(outfile, mesg);

-------------------------------------------------------------------------------
--  	  print_source_message("Source: ", name, " generating tokens with ",
--                                out_control, "ang distribution; Mean & Var: ",
--                                generation.mean, generation.var);
-------------------------------------------------------------------------------

  	  end if;
          out_arc <= (id, sending, TRUE, now);
  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
          write(mesg, string'("Token generated at time: "));
	  write(mesg, now);
	  writeline(outfile, mesg);
	  write(mesg, string'("Total tokens generated by Source: "));
	  write(mesg, name);
	  write(mesg, string'(" at Time: "));
	  write(mesg, now);
	  write(mesg, string'(" = "));
	  write(mesg, num_tok_gen);
	  writeline(outfile, mesg);

  	  print_formatted_message("Token generated at time: ", now);
  	  print_token_message("Total tokens generated by Source ", 
  			      " at Time: ", now, " = ", num_tok_gen);
          print_handshake_message("Source: ", name, " sent handshake signal ",
                 		   out_arc.handshake, " to Queue: ",
                                   out_arc.id, " at Time: ", now); 
  	  end if;
          wait on out_arc until (out_arc.handshake = accepted);
	  idle_time := (now - start_time);
  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
  	  print_handshake_message("Source: ", name, " received handshake signal", 
  				   out_arc.handshake,
                                  " from component: ", out_arc.id,
                                  " at Time: ", now);
  	  end if;
          out_arc <= null;
          wait for 2 fs;
	  idle_time := (idle_time + 2 fs);
          GenRnd(erlan);
          rnd := erlan.rnd;
          trand := rnd * 1.0 fs;
          wait for trand;
          total_time := (now - start_time);
          busy_time := (total_time - idle_time);
	  utilization := (real(busy_time/1 fs) / real(total_time/1 fs));
          if (debug_control = summary) then
	    write(mesg, string'("Utilization for Source: "));
	    write(mesg, name);
	    write(mesg, string'("; U = "));
	    write(mesg, utilization);
	    writeline(outfile, mesg);
	  end if;
	    
	elsif (out_control = expo) then
	  assert false
            report "Generating tokens with exponential distribution"
            severity note;
	  start_time := now;
  	  if (debug_control = trace_tokens or 
  	      debug_control = summary) then 
	    write(mesg, string'("Source: "));
	    write(mesg, name);
	    write(mesg, string'(" generating tokens with EXPONENTIAL distribution; Mean: "));
	    write(mesg, generation.mean);
	    writeline(outfile, mesg);

-------------------------------------------------------------------------------
--  	  print_source_message("Source: ", name, " generating tokens with ",
--                                out_control, "nential distribution; Mean: ",
--                                generation.mean, generation.var);
-------------------------------------------------------------------------------

  	  end if;
          out_arc <= (id, sending, TRUE, now);
  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
          write(mesg, string'("Token generated at time: "));
	  write(mesg, now);
	  writeline(outfile, mesg);
	  write(mesg, string'("Total tokens generated by Source: "));
	  write(mesg, name);
	  write(mesg, string'(" at Time: "));
	  write(mesg, now);
	  write(mesg, string'(" = "));
	  write(mesg, num_tok_gen);
	  writeline(outfile, mesg);

  	  print_formatted_message("Token generated at time: ", now);
  	  print_token_message("Total tokens generated by Source ", 
  			      " at Time: ", now, " = ", num_tok_gen);
          print_handshake_message("Source: ", name, " sent handshake signal ",
                 		   out_arc.handshake, " to Queue: ",
                                   out_arc.id, " at Time: ", now); 
  	  end if;
          wait on out_arc until (out_arc.handshake = accepted);
	  idle_time := (now - start_time);
  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
  	  print_handshake_message("Source: ", name, " received handshake signal", 
  				   out_arc.handshake,
                                  " from component: ", out_arc.id,
                                  " at Time: ", now);
  	  end if;
          out_arc <= null;
          wait for 2 fs;
	  idle_time := (idle_time + 2 fs);
          GenRnd(expon);
          rnd := expon.rnd;
          trand := rnd * 1.0 fs;
          wait for trand;
          total_time := (now - start_time);
          busy_time := (total_time - idle_time);
	  utilization := (real(busy_time/1 fs) / real(total_time/1 fs));
          if (debug_control = summary) then
	    write(mesg, string'("Utilization for Source: "));
	    write(mesg, name);
	    write(mesg, string'("; U = "));
	    write(mesg, utilization);
	    writeline(outfile, mesg);
	  end if;
	    
	elsif (out_control = negex) then
	  assert false
            report "Generating tokens with negative exponential distribution"
            severity note;
	  start_time := now;
  	  if (debug_control = trace_tokens or 
  	      debug_control = summary) then 
	    write(mesg, string'("Source: "));
	    write(mesg, name);
	    write(mesg, string'(" generating tokens with NEG-EXPO distribution; Mean: "));
	    write(mesg, generation.mean);
	    writeline(outfile, mesg);

-------------------------------------------------------------------------------
--  	  print_source_message("Source: ", name, " generating tokens with ",
--                                out_control, "ponential distribution; Mean: ",
--                                generation.mean, generation.var);
-------------------------------------------------------------------------------

  	  end if;
          out_arc <= (id, sending, TRUE, now);
  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
          write(mesg, string'("Token generated at time: "));
	  write(mesg, now);
	  writeline(outfile, mesg);
	  write(mesg, string'("Total tokens generated by Source: "));
	  write(mesg, name);
	  write(mesg, string'(" at Time: "));
	  write(mesg, now);
	  write(mesg, string'(" = "));
	  write(mesg, num_tok_gen);
	  writeline(outfile, mesg);

  	  print_formatted_message("Token generated at time: ", now);
  	  print_token_message("Total tokens generated by Source ", 
  			      " at Time: ", now, " = ", num_tok_gen);
          print_handshake_message("Source: ", name, " sent handshake signal ",
                 		   out_arc.handshake, " to Queue: ",
                                   out_arc.id, " at Time: ", now); 
  	  end if;
          wait on out_arc until (out_arc.handshake = accepted);
	  idle_time := (now - start_time);
  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
  	  print_handshake_message("Source: ", name, " received handshake signal", 
  				   out_arc.handshake,
                                  " from component: ", out_arc.id,
                                  " at Time: ", now);
  	  end if;
          out_arc <= null;
          wait for 2 fs;
	  idle_time := (idle_time + 2 fs);
          GenRnd(nex);
          rnd := nex.rnd;
          trand := rnd * 1.0 fs;
          wait for trand;
          total_time := (now - start_time);
          busy_time := (total_time - idle_time);
	  utilization := (real(busy_time/1 fs) / real(total_time/1 fs));
          if (debug_control = summary) then
	    write(mesg, string'("Utilization for Source: "));
	    write(mesg, name);
	    write(mesg, string'("; U = "));
	    write(mesg, utilization);
	    writeline(outfile, mesg);
	  end if;
	    
	elsif (out_control = fixed) then
	  assert false
            report "Generating tokens with fixed distribution"
            severity note;
	  start_time := now;
  	  if (debug_control = trace_tokens or 
  	      debug_control = summary) then 
	    write(mesg, string'("Source: "));
	    write(mesg, name);
	    write(mesg, string'(" generating tokens at FIXED intervals of "));
	    write(mesg, generation.mean);
	    write(mesg, string'("fs"));
	    writeline(outfile, mesg);

-------------------------------------------------------------------------------
--  	  print_source_message("Source: ", name, " generating tokens with ",
--                                out_control, " distribution; Mean & Var: ",
--                                generation.mean, generation.var);
-------------------------------------------------------------------------------

  	  end if;
          out_arc <= (id, sending, TRUE, now);
  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
          write(mesg, string'("Token generated at time: "));
	  write(mesg, now);
	  writeline(outfile, mesg);
	  write(mesg, string'("Total tokens generated by Source: "));
	  write(mesg, name);
	  write(mesg, string'(" at Time: "));
	  write(mesg, now);
	  write(mesg, string'(" = "));
	  write(mesg, num_tok_gen);
	  writeline(outfile, mesg);

  	  print_formatted_message("Token generated at time: ", now);
  	  print_token_message("Total tokens generated by Source ", 
  			      " at Time: ", now, " = ", num_tok_gen);
  	  print_handshake_message("Source: ", name, " sent handshake signal ",
                		   out_arc.handshake, " to Queue: ",
                                   out_arc.id, " at Time: ", now);
  	  end if;
          wait on out_arc until (out_arc.handshake = accepted);
	  idle_time := (now - start_time);
  	  if (debug_control = handshaking or 
  	      debug_control = summary) then 
  	  print_handshake_message("Source: ", name, " received handshake signal", 
  				   out_arc.handshake,
                                  " from component: ", out_arc.id,
                                  " at Time: ", now);
  	  end if;
          out_arc <= null;
          wait for 2 fs;
	  idle_time := (idle_time + 2 fs);
          trand := generation.mean * 1.0 fs;
          wait for trand;
          total_time := (now - start_time);
          busy_time := (total_time - idle_time);
	  utilization := (real(busy_time/1 fs) / real(total_time/1 fs));
          if (debug_control = summary) then
	    write(mesg, string'("Utilization for Source: "));
	    write(mesg, name);
	    write(mesg, string'("; U = "));
	    write(mesg, utilization);
	    writeline(outfile, mesg);
	  end if;

         else assert false
		report "Out_control value wrongly specified"
		severity error;
        end if;

    end loop;

end process;

end behavior;
